#include "channel_element.h"
#include "file_manager.h"

ChannelElement::ChannelElement()
{
  // Initialization will be done in runForever.
  thisProxy[thisIndex].runForever();
}

ChannelElement::ChannelElement(CkMigrateMessage* msg)
{
  // Do nothing.
}

bool ChannelElement::allInitialized()
{
  int  edge;               // Loop counter.
  bool initialized = true; // Flag to record whether we have found an uninitialized neighbor.
  
  for (edge = 0; initialized && edge < channelNeighborsSize; edge++)
    {
      initialized = channelNeighborsInitialized[edge];
    }
  
  for (edge = 0; initialized && edge < meshNeighborsSize; edge++)
    {
      initialized = meshNeighborsInitialized[edge];
    }
  
  return initialized;
}

void ChannelElement::handleInitialize(CProxy_MeshElement meshProxyInit, CProxy_FileManager fileManagerProxyInit)
{
  bool         error                  = false;                                                        // Error flag.
  int          edge;                                                                                  // Loop counter.
  FileManager* fileManagerLocalBranch = fileManagerProxyInit.ckLocalBranch();                         // Used for access to local public member variables.
  int          fileManagerLocalIndex  = thisIndex - fileManagerLocalBranch->localChannelElementStart; // Index of this element in file manager arrays.
  
      meshProxy        = meshProxyInit;
      fileManagerProxy = fileManagerProxyInit;
  
  if (!error)
    {
      groundwaterDone          = true;
      surfacewaterDone         = true;
      dt                       = 1.0;
      dtNew                    = 1.0;
      
      if (NULL != fileManagerLocalBranch->channelChannelNeighbors)
        {
          for (edge = 0; !error && edge < channelNeighborsSize; edge++)
            {
              channelNeighbors[edge]                           = fileManagerLocalBranch->channelChannelNeighbors[fileManagerLocalIndex][edge];
              channelNeighborsInvariantChecked[edge]           = true;
              channelNeighborsSurfacewaterFlowRateReady[edge]  = FLOW_RATE_LIMITING_CHECK_DONE;
              channelNeighborsSurfacewaterFlowRate[edge]       = 0.0;
              channelNeighborsSurfacewaterCumulativeFlow[edge] = 0.0;
              
              if (0 == edge)
                {
                  channelNeighbors[edge] = (thisIndex - 1 + fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements) % fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements;
                }
              else if (1 == edge)
                {
                  channelNeighbors[edge] = (thisIndex + 1) % fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements;
                }
              else
                {
                  channelNeighbors[edge] = NOFLOW;
                }
              
              if (isBoundary(channelNeighbors[edge]))
                {
                  channelNeighborsInitialized[edge] = true;
                  
                  // Unused, but initialize for completeness.
                  channelNeighborsReciprocalEdge[edge] = 0;
                  channelNeighborsInteraction[edge]    = BOTH_CALCULATE_FLOW_RATE;
                  channelNeighborsZBank[edge]          = 0.0;
                  channelNeighborsZBed[edge]           = 0.0;
                  channelNeighborsLength[edge]         = 1.0;
                  channelNeighborsChannelType[edge]    = STREAM;
                  channelNeighborsBaseWidth[edge]      = 1.0;
                  channelNeighborsSideSlope[edge]      = 1.0;
                  channelNeighborsManningsN[edge]      = 1.0;
                }
              else if (0 <= channelNeighbors[edge] && channelNeighbors[edge] < fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements)
                {
                  channelNeighborsInitialized[edge] = false;
                  
                  thisProxy[channelNeighbors[edge]].initializeChannelNeighbor(thisIndex, edge, elementZBank, elementZBed, elementLength, channelType,
                                                                              baseWidth, sideSlope, manningsN);
                }
              else
                {
                  // We have to finish initialization before checking the invariant so the invariant hasn't been checked yet so we have to check this here.
                  CkError("ERROR in ChannelElement::handleInitialize, element %d, edge %d: channelNeighbors must be a boundary condition code or a valid "
                          "array index.\n", thisIndex, edge);
                  error = true;
                }
            }
        }
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
      else
        {
          CkError("ERROR in ChannelElement::handleInitialize, element %d: channelNeighbors initialization information not available from local file "
                  "manager.\n", thisIndex);
          error = true;
        }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
    }
  
  if (!error)
    {
      if (NULL != fileManagerLocalBranch->channelMeshNeighbors)
        {
          for (edge = 0; !error && edge < meshNeighborsSize; edge++)
            {
              meshNeighbors[edge]                           = fileManagerLocalBranch->channelMeshNeighbors[fileManagerLocalIndex][edge];
              meshNeighborsInvariantChecked[edge]           = true;
              meshNeighborsSurfacewaterFlowRateReady[edge]  = FLOW_RATE_LIMITING_CHECK_DONE;
              meshNeighborsSurfacewaterFlowRate[edge]       = 0.0;
              meshNeighborsSurfacewaterCumulativeFlow[edge] = 0.0;
              meshNeighborsGroundwaterFlowRateReady[edge]   = FLOW_RATE_LIMITING_CHECK_DONE;
              meshNeighborsGroundwaterFlowRate[edge]        = 0.0;
              meshNeighborsGroundwaterCumulativeFlow[edge]  = 0.0;
              
              /* FIXME the error doesn't occur if I add this.
              if (15 > edge)
                {
                  meshNeighbors[edge] = (thisIndex - 7 + edge + fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements) % fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements;
                }
              else
                {
                  meshNeighbors[edge] = NOFLOW;
                }
              */
              
              if (isBoundary(meshNeighbors[edge]))
                {
                  meshNeighborsInitialized[edge] = true;
                  
                  // Unused, but initialize for completeness.
                  meshNeighborsReciprocalEdge[edge] = 0;
                  meshNeighborsInteraction[edge]    = BOTH_CALCULATE_FLOW_RATE;
                  meshNeighborsZSurface[edge]       = 0.0;
                  meshNeighborsZBedrock[edge]       = 0.0;
                  meshNeighborsZOffset[edge]        = 0.0;
                }
              else if (0 <= meshNeighbors[edge] && meshNeighbors[edge] < fileManagerProxy.ckLocalBranch()->globalNumberOfMeshElements)
                {
                  meshNeighborsInitialized[edge] = false;
                  
                  meshProxy[meshNeighbors[edge]].initializeChannelNeighbor(thisIndex, edge, elementX, elementY, elementZBank, elementZBed, baseWidth,
                                                                           sideSlope, bedConductivity, bedThickness);
                }
              else
                {
                  // We have to finish initialization before checking the invariant so the invariant hasn't been checked yet so we have to check this here.
                  CkError("ERROR in ChannelElement::handleInitialize, element %d, edge %d: meshNeighbors must be a boundary condition code or a valid "
                          "array index.\n", thisIndex, edge);
                  error = true;
                }
            }
        }
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
      else
        {
          CkError("ERROR in ChannelElement::handleInitialize, element %d: meshNeighbors initialization information not available from local file manager.\n",
                  thisIndex);
          error = true;
        }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
    }
  
  if (error)
    {
      CkExit();
    }
}

void ChannelElement::handleInitializeChannelNeighbor(int neighbor, int neighborReciprocalEdge, double neighborZBank, double neighborZBed,
                                                     double neighborLength, ChannelTypeEnum neighborChannelType, double neighborBaseWidth,
                                                     double neighborSideSlope, double neighborManningsN)
{
  bool error = false; // Error flag.
  int  edge  = 0;     // Loop counter.
  
  while (edge < channelNeighborsSize - 1 && channelNeighbors[edge] != neighbor)
    {
      edge++;
    }
  
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  if (!(channelNeighbors[edge] == neighbor))
    {
      CkError("ERROR in ChannelElement::handleInitializeChannelNeighbor, element %d: received an initialization message from an element that is not my "
              "neighbor.\n", thisIndex);
      error = true;
    }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  
  if (!error)
    {
      channelNeighborsInitialized[edge]     = true;
      channelNeighborsReciprocalEdge[edge]  = neighborReciprocalEdge;
      channelNeighborsInteraction[edge]     = BOTH_CALCULATE_FLOW_RATE;
      channelNeighborsZBank[edge]           = neighborZBank;
      channelNeighborsZBed[edge]            = neighborZBed;
      channelNeighborsLength[edge]          = neighborLength;
      channelNeighborsChannelType[edge]     = neighborChannelType;
      channelNeighborsBaseWidth[edge]       = neighborBaseWidth;
      channelNeighborsSideSlope[edge]       = neighborSideSlope;
      channelNeighborsManningsN[edge]       = neighborManningsN;
    }
  
  // Error checking of initialization values is done in the invariant.
  
  if (error)
    {
      CkExit();
    }
}

void ChannelElement::handleInitializeMeshNeighbor(int neighbor, int neighborReciprocalEdge, double neighborX, double neighborY, double neighborZSurface,
                                                  double neighborZBedrock, double neighborSlopeX, double neighborSlopeY)
{
  bool error = false; // Error flag.
  int  edge  = 0;     // Loop counter.
  
  while (edge < meshNeighborsSize - 1 && meshNeighbors[edge] != neighbor)
    {
      edge++;
    }
  
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  if (!(meshNeighbors[edge] == neighbor))
    {
      CkError("ERROR in ChannelElement::handleInitializeMeshNeighbor, element %d: received an initialization message from an element that is not my "
              "neighbor.\n", thisIndex);
      error = true;
    }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  
  if (!error)
    {
      meshNeighborsInitialized[edge]    = true;
      meshNeighborsReciprocalEdge[edge] = neighborReciprocalEdge;
      meshNeighborsInteraction[edge]    = BOTH_CALCULATE_FLOW_RATE;
      meshNeighborsZSurface[edge]       = neighborZSurface;
      meshNeighborsZBedrock[edge]       = neighborZBedrock;
      meshNeighborsZOffset[edge]        = (elementX - neighborX) * neighborSlopeX + (elementY - neighborY) * neighborSlopeY;
    }
  
  // Error checking of initialization values is done in the invariant.
  
  if (error)
    {
      CkExit();
    }
}

void ChannelElement::handleForcingDataMessage()
{
  // FIXME add forcing data for channels.
  
  contribute();
}

// Suppress warnings in the The Charm++ autogenerated code.
#pragma GCC diagnostic ignored "-Wunused-variable"
#include "channel_element.def.h"
#pragma GCC diagnostic warning "-Wunused-variable"
