#include "mesh_element.h"
#include "file_manager.h"

MeshElement::MeshElement()
{
  // Initialization will be done in runForever.
  thisProxy[thisIndex].runForever();
}

MeshElement::MeshElement(CkMigrateMessage* msg)
{
  // Do nothing.
}

bool MeshElement::allInitialized()
{
  int  edge;               // Loop counter.
  bool initialized = true; // Flag to record whether we have found an uninitialized neighbor.
  
  for (edge = 0; initialized && edge < meshNeighborsSize; edge++)
    {
      initialized = meshNeighborsInitialized[edge];
    }
  
  for (edge = 0; initialized && edge < channelNeighborsSize; edge++)
    {
      initialized = channelNeighborsInitialized[edge];
    }
  
  return initialized;
}

void MeshElement::handleInitialize(CProxy_ChannelElement channelProxyInit, CProxy_FileManager fileManagerProxyInit)
{
  bool         error                  = false;                                                     // Error flag.
  int          edge;                                                                               // Loop counter.
  FileManager* fileManagerLocalBranch = fileManagerProxyInit.ckLocalBranch();                      // Used for access to local public member variables.
  int          fileManagerLocalIndex  = thisIndex - fileManagerLocalBranch->localMeshElementStart; // Index of this element in file manager arrays.
  
      channelProxy     = channelProxyInit;
      fileManagerProxy = fileManagerProxyInit;
  
  if (!error)
    {
      precipitation            = 0.0;
      precipitationCumulative  = 0.0;
      evaporation              = 0.0;
      evaporationCumulative    = 0.0;
      surfacewaterInfiltration = 0.0;
      groundwaterRecharge      = 0.0;
      groundwaterDone          = true;
      infiltrationDone         = true;
      surfacewaterDone         = true;
      dt                       = 1.0;
      dtNew                    = 1.0;
      
      if (NULL != fileManagerLocalBranch->meshMeshNeighbors)
        {
          for (edge = 0; !error && edge < meshNeighborsSize; edge++)
            {
              meshNeighbors[edge]                           = fileManagerLocalBranch->meshMeshNeighbors[fileManagerLocalIndex][edge];
              meshNeighborsChannelEdge[edge]                = false;
              meshNeighborsInvariantChecked[edge]           = true;
              meshNeighborsSurfacewaterFlowRateReady[edge]  = FLOW_RATE_LIMITING_CHECK_DONE;
              meshNeighborsSurfacewaterFlowRate[edge]       = 0.0;
              meshNeighborsSurfacewaterCumulativeFlow[edge] = 0.0;
              meshNeighborsGroundwaterFlowRateReady[edge]   = FLOW_RATE_LIMITING_CHECK_DONE;
              meshNeighborsGroundwaterFlowRate[edge]        = 0.0;
              meshNeighborsGroundwaterCumulativeFlow[edge]  = 0.0;
              
              if (0 == edge)
                {
                  meshNeighbors[edge] = (thisIndex - 1 + fileManagerProxy.ckLocalBranch()->globalNumberOfMeshElements) % fileManagerProxy.ckLocalBranch()->globalNumberOfMeshElements;
                }
              else if (1 == edge)
                {
                  meshNeighbors[edge] = (thisIndex + 1) % fileManagerProxy.ckLocalBranch()->globalNumberOfMeshElements;
                }
              else
                {
                  meshNeighbors[edge] = NOFLOW;
                }
              
              if (isBoundary(meshNeighbors[edge]))
                {
                  meshNeighborsInitialized[edge] = true;
                  
                  // Unused, but initialize for completeness.
                  meshNeighborsReciprocalEdge[edge] = 0;
                  meshNeighborsInteraction[edge]    = BOTH_CALCULATE_FLOW_RATE;
                  meshNeighborsX[edge]              = 0.0;
                  meshNeighborsY[edge]              = 0.0;
                  meshNeighborsZSurface[edge]       = 0.0;
                  meshNeighborsZBedrock[edge]       = 0.0;
                  meshNeighborsArea[edge]           = 1.0;
                  meshNeighborsConductivity[edge]   = 1.0;
                  meshNeighborsManningsN[edge]      = 1.0;
                }
              else if (0 <= meshNeighbors[edge] && meshNeighbors[edge] < fileManagerProxy.ckLocalBranch()->globalNumberOfMeshElements)
                {
                  meshNeighborsInitialized[edge] = false;
                  
                  thisProxy[meshNeighbors[edge]].initializeMeshNeighbor(thisIndex, edge, elementX, elementY, elementZSurface, elementZBedrock, elementArea,
                                                                        conductivity, manningsN);
                }
              else
                {
                  // We have to finish initialization before checking the invariant so the invariant hasn't been checked yet so we have to check this here.
                  CkError("ERROR in MeshElement::handleInitialize, element %d, edge %d: meshNeighbors must be a boundary condition code or a valid "
                          "array index.\n", thisIndex, edge);
                  error = true;
                }
            }
        }
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
      else
        {
          CkError("ERROR in MeshElement::handleInitialize, element %d: meshNeighbors initialization information not available from local file manager.\n",
                  thisIndex);
          error = true;
        }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
    }
  
  if (!error)
    {
      if (NULL != fileManagerLocalBranch->meshChannelNeighbors)
        {
          for (edge = 0; !error && edge < channelNeighborsSize; edge++)
            {
              channelNeighbors[edge]                           = fileManagerLocalBranch->meshChannelNeighbors[fileManagerLocalIndex][edge];
              channelNeighborsInvariantChecked[edge]           = true;
              channelNeighborsSurfacewaterFlowRateReady[edge]  = FLOW_RATE_LIMITING_CHECK_DONE;
              channelNeighborsSurfacewaterFlowRate[edge]       = 0.0;
              channelNeighborsSurfacewaterCumulativeFlow[edge] = 0.0;
              channelNeighborsGroundwaterFlowRateReady[edge]   = FLOW_RATE_LIMITING_CHECK_DONE;
              channelNeighborsGroundwaterFlowRate[edge]        = 0.0;
              channelNeighborsGroundwaterCumulativeFlow[edge]  = 0.0;
              
              /* FIXME the error doesn't occur if I add this.
              if (thisIndex < fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements && 15 > edge)
                {
                  channelNeighbors[edge] = (thisIndex - 7 + edge + fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements) % fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements;
                }
              else
                {
                  channelNeighbors[edge] = NOFLOW;
                }
                */
              
              if (isBoundary(channelNeighbors[edge]))
                {
                  channelNeighborsInitialized[edge] = true;
                  
                  // Unused, but initialize for completeness.
                  channelNeighborsReciprocalEdge[edge]  = 0;
                  channelNeighborsInteraction[edge]     = BOTH_CALCULATE_FLOW_RATE;
                  channelNeighborsZBank[edge]           = 0.0;
                  channelNeighborsZBed[edge]            = 0.0;
                  channelNeighborsZOffset[edge]         = 0.0;
                  channelNeighborsBaseWidth[edge]       = 1.0;
                  channelNeighborsSideSlope[edge]       = 1.0;
                  channelNeighborsBedConductivity[edge] = 1.0;
                  channelNeighborsBedThickness[edge]    = 1.0;
                }
              else if (0 <= channelNeighbors[edge] && channelNeighbors[edge] < fileManagerProxy.ckLocalBranch()->globalNumberOfChannelElements)
                {
                  channelNeighborsInitialized[edge] = false;
                  
                  channelProxy[channelNeighbors[edge]].initializeMeshNeighbor(thisIndex, edge, elementX, elementY, elementZSurface, elementZBedrock,
                                                                              elementSlopeX, elementSlopeY);
                }
              else
                {
                  // We have to finish initialization before checking the invariant so the invariant hasn't been checked yet so we have to check this here.
                  CkError("ERROR in MeshElement::handleInitialize, element %d, edge %d: channelNeighbors must be a boundary condition code or a valid "
                          "array index.\n", thisIndex, edge);
                  error = true;
                }
            }
        }
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
      else
        {
          CkError("ERROR in MeshElement::handleInitialize, element %d: channelNeighbors initialization information not available from local file manager.\n",
                  thisIndex);
          error = true;
        }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
    }
  
  if (error)
    {
      CkExit();
    }
}

void MeshElement::handleInitializeMeshNeighbor(int neighbor, int neighborReciprocalEdge, double neighborX, double neighborY, double neighborZSurface,
                                               double neighborZBedrock, double neighborArea, double neighborConductivity, double neighborManningsN)
{
  bool error = false; // Error flag.
  int  edge  = 0;     // Loop counter.
  
  while (edge < meshNeighborsSize - 1 && meshNeighbors[edge] != neighbor)
    {
      edge++;
    }
  
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  if (!(meshNeighbors[edge] == neighbor))
    {
      CkError("ERROR in MeshElement::handleInitializeMeshNeighbor, element %d: received an initialization message from an element that is not my neighbor.\n",
              thisIndex);
      error = true;
    }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  
  if (!error)
    {
      meshNeighborsInitialized[edge]    = true;
      meshNeighborsReciprocalEdge[edge] = neighborReciprocalEdge;
      meshNeighborsInteraction[edge]    = BOTH_CALCULATE_FLOW_RATE;
      meshNeighborsX[edge]              = neighborX;
      meshNeighborsY[edge]              = neighborY;
      meshNeighborsZSurface[edge]       = neighborZSurface;
      meshNeighborsZBedrock[edge]       = neighborZBedrock;
      meshNeighborsArea[edge]           = neighborArea;
      meshNeighborsConductivity[edge]   = neighborConductivity;
      meshNeighborsManningsN[edge]      = neighborManningsN;
    }
  
  // Error checking of initialization values is done in the invariant.
  
  if (error)
    {
      CkExit();
    }
}

void MeshElement::handleInitializeChannelNeighbor(int neighbor, int neighborReciprocalEdge, double neighborX, double neighborY, double neighborZBank,
                                                  double neighborZBed, double neighborBaseWidth, double neighborSideSlope, double neighborBedConductivity,
                                                  double neighborBedThickness)
{
  bool error = false; // Error flag.
  int  edge  = 0;     // Loop counter.
  
  while (edge < channelNeighborsSize - 1 && channelNeighbors[edge] != neighbor)
    {
      edge++;
    }
  
#if (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  if (!(channelNeighbors[edge] == neighbor))
    {
      CkError("ERROR in MeshElement::handleInitializeChannelNeighbor, element %d: received an initialization message from an element that is not my "
              "neighbor.\n", thisIndex);
      error = true;
    }
#endif // (DEBUG_LEVEL & DEBUG_LEVEL_USER_INPUT_SIMPLE)
  
  if (!error)
    {
      channelNeighborsInitialized[edge]     = true;
      channelNeighborsReciprocalEdge[edge]  = neighborReciprocalEdge;
      channelNeighborsInteraction[edge]     = BOTH_CALCULATE_FLOW_RATE;
      channelNeighborsZBank[edge]           = neighborZBank;
      channelNeighborsZBed[edge]            = neighborZBed;
      channelNeighborsZOffset[edge]         = (neighborX - elementX) * elementSlopeX + (neighborY - elementY) * elementSlopeY;
      channelNeighborsBaseWidth[edge]       = neighborBaseWidth;
      channelNeighborsSideSlope[edge]       = neighborSideSlope;
      channelNeighborsBedConductivity[edge] = neighborBedConductivity;
      channelNeighborsBedThickness[edge]    = neighborBedThickness;
    }
  
  // Error checking of initialization values is done in the invariant.
  
  if (error)
    {
      CkExit();
    }
}

void MeshElement::handleForcingDataMessage(double atmosphereLayerThicknessNew, double shadedFractionNew, double shadedFractionMaximumNew,
                                           double surfaceTemperatureNew, double surfacePressureNew, double atomsphereLayerPressureNew, double eastWindSpeedNew,
                                           double northWindSpeedNew, double atmosphereLayerMixingRatioNew, double cloudMixingRatioNew,
                                           double shortWaveRadiationDownNew, double longWaveRadiationDownNew, double precipitationRateNew,
                                           double soilBottomTemperatureNew, double planetaryBoundaryLayerHeightNew)
{
  // FIXME error checking on inputs.
  
  atmosphereLayerThickness     = atmosphereLayerThicknessNew;
  shadedFraction               = shadedFractionNew;
  shadedFractionMaximum        = shadedFractionMaximumNew;
  surfaceTemperature           = surfaceTemperatureNew;
  surfacePressure              = surfacePressureNew;
  atomsphereLayerPressure      = atomsphereLayerPressureNew;
  eastWindSpeed                = eastWindSpeedNew;
  northWindSpeed               = northWindSpeedNew;
  atmosphereLayerMixingRatio   = atmosphereLayerMixingRatioNew;
  cloudMixingRatio             = cloudMixingRatioNew;
  shortWaveRadiationDown       = shortWaveRadiationDownNew;
  longWaveRadiationDown        = longWaveRadiationDownNew;
  precipitationRate            = precipitationRateNew;
  soilBottomTemperature        = soilBottomTemperatureNew;
  planetaryBoundaryLayerHeight = planetaryBoundaryLayerHeightNew;
  
  contribute();
}

// Suppress warnings in the The Charm++ autogenerated code.
#pragma GCC diagnostic ignored "-Wunused-variable"
#include "mesh_element.def.h"
#pragma GCC diagnostic warning "-Wunused-variable"
